;;;			    mew-header.el
;;;
;;;	      Copyright (C) 1996-1997  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazu Yamamoto <Kazu@Mew.org>
;;; Created: Oct  2, 1996
;;; Revised: Aug 20, 1997
;;;

(defconst mew-header-version "mew-header.el version 0.12")

(require 'mew)

(defvar mew-header-max-depth 20
  "*A value to decide loop depth for header field syntax analysis.
It was known as mew-loop-depth.
See also mew-header-max-length.")

(defvar mew-header-decode-regex 
  "=\\?\\([^?]+\\)\\?\\(.\\)\\?\\([^?]+\\)\\?=")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Header functions
;;;

(defun mew-header-get-value (field)
  "currently, when no match, it returns nil."
  (let ((case-fold-search t)
	(start nil)
	(key nil)
	(match nil)
	(count 0)
	(ret nil)) ;; (concat nil "foo") -> "foo"
    (save-excursion
      (goto-char (point-min))
      (catch 'header
	(while (re-search-forward (format "^-*$\\|^%s[ \t]*" field) nil t)
	  (if (>= count mew-header-max-depth)
	      (throw 'header ret)) ;; xxx warn to users?
	  (setq count (1+ count))
	  (setq key (mew-match 0))
	  (setq start (match-end 0))
	  (if (string-match "^-*$" key) (throw 'header ret))
	  (forward-line)
	  (while (looking-at mew-lwsp) (forward-line))
	  (setq match (mew-buffer-substring start (1- (point))))
	  (if (null ret)
	      (if (null (string= "" match))
		  (setq ret match))
	    (if (null (string= "" match))
		(setq ret (concat ret "," match))))
	)))
    ret
    ))
	  
(defun mew-header-get-line (field)
;; return  with ^J
;; currently, when no match, it returns nil.
  (let ((case-fold-search t)
	(start nil)
	(key nil)
	(count 0)
	(ret nil)) ;; (concat nil "foo") -> "foo"
    (save-excursion
      (goto-char (point-min))
      (catch 'header
	(while (re-search-forward (format "^-*$\\|^%s" field) nil t)
	  (if (>= count mew-header-max-depth)
	      (throw 'header ret)) ;; xxx warn to users?
	  (setq count (1+ count))
	  (setq key (mew-match 0))
	  (setq start (match-beginning 0))
	  (if (string-match "^-*$" key) (throw 'header ret))
	  (forward-line)
	  (while (looking-at mew-lwsp) (forward-line))
	  (setq ret (concat ret (mew-buffer-substring start (point))))
	  ))
      )
    ret
    ))

(defun mew-header-delete-lines (fields)
  (let ((case-fold-search t)
	(regex (mew-make-field-regex fields))
	key start)
    (save-excursion
      (goto-char (point-min))
      (catch 'header
	(while (re-search-forward (concat "^-*$\\|" regex) nil t)
	  (setq key (mew-match 0)
		start (match-beginning 0))
	  (if (string-match "^-*$" key) (throw 'header nil))
	  (forward-line)
	  (while (looking-at mew-lwsp) (forward-line))
	  (delete-region start (point))
	  ))
      )))

(defun mew-header-resent-lines (fields)
  (let ((case-fold-search t)
	(regex (mew-make-field-regex fields))
	key)
    (save-excursion
      (goto-char (point-min))
      (catch 'header
	(while (re-search-forward (concat "^-*$\\|" regex) nil t)
	  (setq key (mew-match 0))
	  (if (string-match "^-*$" key) (throw 'header nil))
	  (beginning-of-line)
	  (insert "Prev-")
	  ))
      )))

(defun mew-header-insert-here (field &optional value noret)
  (insert field)
  (insert " ")
  (if value (insert value))
  (if (not noret) (insert "\n"))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Low-level functions to parse fields
;;;

;;
;; Get a canonical syntax
;;
;; '"Winnie (The) Pooh" (loves "honey") @ (somewhere in) England'
;; ->
;; '"Winnie (The) Pooh"@England'
;;
(defun mew-header-syntax (string)
  (let* ((i 0) (len (length string))
	 (par-cnt 0) (ret-cnt 0)
	 (ret (make-string len ?x))
	 quoted c)
    (while (< i len)
      (setq c (aref string i))
      (cond
       (quoted
	(if (char-equal c ?\") (setq quoted nil))
	(aset ret ret-cnt c)
	(setq ret-cnt (1+ ret-cnt))
	)
       ((> par-cnt 0)
	(cond
	 ((char-equal c ?\() (setq par-cnt (1+ par-cnt)))
	 ((char-equal c ?\)) (setq par-cnt (1- par-cnt)))
	 )
	)
       ((char-equal c ?\() (setq par-cnt 1))
       ((char-equal c ?\n))
       ((char-equal c ?\t))
       ((char-equal c ?\ ))
       (t 
	(if (char-equal c ?\") (setq quoted t))
	(aset ret ret-cnt c)
	(setq ret-cnt (1+ ret-cnt))
	)
       )
      (setq i (1+ i))
      )
    (substring ret 0 ret-cnt)
    ))

;;
;; Split a string by a separator.
;;
;; "Winnie-The-Pooh" -> "Winnie", "The", "Pooh"
;;
(defun mew-header-split (string sep)
 (let ((quote nil)
       (len (length string))
       (ret nil)
       (start 0)
       (n 0))
   (while (< n len)
     (cond
      ((char-equal (aref string n) ?\")
       (setq quote (not quote)))
      ((char-equal (aref string n) sep)
       (if (null quote)
	   (progn
	     (setq ret (cons (substring string start n) ret))
	     (setq start (1+ n))
	     )
	 ))
      )
     (setq n (1+ n))
     )
   (setq ret (cons (substring string start n) ret))
   (nreverse ret)
   ))

;;
;; Extract an address from <>.
;;
;; "Winnie The Pooh <wtp@uk>" -> "wtp@uk"
;;
(defmacro mew-header-extract-addr-list (list)
  (` (mapcar (function mew-header-extract-addr) (, list))))

(defun mew-header-extract-addr (str)
  "Extracts a real e-mail address from STR and returns it.
e.g. \"Mine Sakurai <m-sakura@ccs.mt.nec.co.jp>\"
  ->  \"m-sakura@ccs.mt.nec.co.jp\".
e.g. \"m-sakura@ccs.mt.nec.co.jp (Mine Sakurai)\"
  ->  \"m-sakura@ccs.mt.nec.co.jp\"."
  (cond ((string-match ".*<\\([^>]*\\)>" str) ;; .* to extract last <>
         (mew-match 1 str))
        ((string-match "\\([^ \t]*@[^ \t]*\\)" str)
         (mew-match 1 str))
        (t str)
    )
  )

;;
;; Delete a string after @.
;;
;; "Pooh@England" -> "Pooh"
;;
(defmacro mew-header-delete-at-list (list)
  (` (mapcar (function mew-header-delete-at) (, list))))

(defun mew-header-delete-at (string)
  (if (string-match "@.*:" string)
      (setq string (substring string (match-end 0) (length string)))
    (setq string (substring string 0 (string-match "%" string)))
    (setq string (substring string 0 (string-match "@" string)))
    (substring string 0 (string-match ":;" string))
    ))

;;
;; Delete null strings.
;;
;; "Winnie", "", "The", "", "Pooh"
;; ->
;; "Winnie", "The", "Pooh"
;;
(defun mew-header-delete-nullstring-list (list)
  (if (null list)
      ()
    (cond 
     ((string= (car list) "") 
      (mew-header-delete-nullstring-list (cdr list)))
     (t 
      (cons (car list) (mew-header-delete-nullstring-list (cdr list)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; High-level functions to parse fields
;;;

(defun mew-header-syntax-list (string)
  (mew-header-split
   (mew-header-syntax string)
   ?;))

(defun mew-header-user-collect (list)
  (let ((fields (mapcar (function mew-header-get-value) list))
	(users "") (count 0) (begin 0))
    (while fields
      (if (car fields) 
	  (setq users (concat users "," (car fields))))
      (setq fields (cdr fields)))
    (while (and (< count mew-header-max-depth) 
		(string-match "," users begin))
      (setq begin (match-end 0))
      (setq count (1+ count))
      )
    (if (equal count mew-header-max-depth)
	(setq users (substring users 0 begin)))
    (mew-header-delete-nullstring-list
     (mew-header-delete-at-list
      (mew-header-extract-addr-list
       (mew-header-split
	(mew-header-syntax users)
	?,))))
    ))


(defun mew-header-address-collect (list)
  (let ((fields (mapcar (function mew-header-get-value) list))
	(users "") (count 0) (begin 0))
    (while fields
      (if (car fields) 
	  (setq users (concat users "," (car fields))))
      (setq fields (cdr fields)))
    (while (and (< count mew-header-max-depth) 
		(string-match "," users begin))
      (setq begin (match-end 0))
      (setq count (1+ count))
      )
    (if (>= count mew-header-max-depth)
	(progn
	  (message "Too many addresses, truncated after %d" 
		   mew-header-max-depth)
	  (ding)
	  (sit-for 3) ;; enough time to read?
	  ))
    (if (equal count mew-header-max-depth) 
	(setq users (substring users 0 begin)))
    (mew-header-delete-nullstring-list
     (mew-header-extract-addr-list
      (mew-header-split
       (mew-header-syntax users)
       ?,)))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Alias functions
;;;

(defun mew-header-expand-alias (alias)
  (let ((assoc (mew-assoc-case-equal alias mew-alias-alist 0)))
    (if assoc 
	(cdr assoc)
      alias)
    ))

(defun mew-header-expand-alias-list (alias-list)
  (mapcar (function mew-header-expand-alias) alias-list)
  )

(defun mew-header-decrypters-collect (list)
  (let ((fields (mapcar (function mew-header-get-value) list))
	(users "") (count 0) (begin 0) a)
    (catch 'first
      (while fields
	(setq a (car fields) fields (cdr fields))
	(if (null a)
	    ()
	  (setq users a)
	  (throw 'first nil)
	  )))
    (while fields
      (if (car fields) 
	  (setq users (concat users "," (car fields))))
      (setq fields (cdr fields)))
    (while (and (< count mew-header-max-depth) 
		(string-match "," users begin))
      (setq begin (match-end 0))
      (setq count (1+ count))
      )
    (if (>= count mew-header-max-depth)
	(progn
	  (message "Too many addresses, truncated after %d" 
		   mew-header-max-depth)
	  (ding)
	  (sit-for 3) ;; enough time to read?
	  ))
    (if (equal count mew-header-max-depth) 
	(setq users (substring users 0 begin)))
    users
    ))

(defun mew-header-get-address (field)
  (let ((value (mew-header-get-value field)))
    (if (null value)
	nil
      (mew-header-extract-addr (mew-header-syntax value))
      )
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; e-mail address canonical form 
;;;

(defmacro mew-header-canform-list (list)
  (` (mapcar (function mew-header-canform) (, list))))

(defun mew-header-canform (string)
  "Complete STRING with mew-mail-domain."
  (if (or (null mew-mail-domain) (string= mew-mail-domain ""))
      string
    (if (string-match "@" string)
	string
      (concat string "@" mew-mail-domain)))
  )

(provide 'mew-header)
