;;;			     mew-mark.el
;;;
;;;		Copyright (C) 1997  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazu Yamamoto <Kazu@Mew.org>
;;; Created: Mar  2, 1997
;;; Revised: Oct 25, 1997
;;;

(defconst mew-mark-version "mew-mark.el version 0.10")

(require 'mew)

;; level 1: *, @
;; level 2: D
;; level 3: o

(defvar mew-mark-multi  ?@)
(defvar mew-mark-review ?*)
(defvar mew-mark-delete ?D)
(defvar mew-mark-refile ?o) ;; do you like "^"?

(defvar mew-mark-tmp    ?%) ;; temporary use only.

(defvar mew-mark-delete-switch
  (list 
   (cons mew-mark-multi  'mew-mark-keep-line)
   (cons mew-mark-review 'mew-mark-keep-line)
   (cons mew-mark-delete 'mew-mark-delete-line)
   (cons mew-mark-refile 'mew-mark-delete-line)
   )
  )

(defvar mew-mark-undo-switch
  (list 
   (cons mew-mark-multi  'mew-mark-unmark)
   (cons mew-mark-review 'mew-mark-unmark)
   (cons mew-mark-delete 'mew-mark-unmark)
   (cons mew-mark-refile 'mew-mark-unrefile)
   )
  )

;;
;; for mew-mark-delete-switch
;;

(defun mew-mark-delete-line ()
  (beginning-of-line)
  (let ((start (point)))
    (forward-line)
    (delete-region start (point))
    ))

(defun mew-mark-keep-line ()
  () ;; do nothing
  )

;;
;; for mew-mark-undo-switch
;;

(defun mew-mark-unmark ()
  "Delete the mark on this message anyway"
  (save-excursion
    (let ((buffer-read-only nil))
      (beginning-of-line)
      (re-search-forward mew-summary-message-regex)
      (delete-char 1)
      (insert " ")
      (mew-highlight-unmark-line)
      )
    ))

(defun mew-mark-unrefile ()
  "Delete refile state and delete the mark"
  (mew-refile-reset (mew-summary-message-number))
  (mew-mark-unmark)
  )

;;
;; Basic functions for mark
;;

(defun mew-summary-marked-p ()
  "See if this message is marked"
  (save-excursion
    (beginning-of-line)
    (cond 
     ((looking-at (concat mew-summary-message-regex " \\|^$")) nil)
     (t t)
     )
    ))

(defun mew-summary-get-mark ()
  (let (mark)
    (save-excursion
      (beginning-of-line)
      (if (not (looking-at (concat mew-summary-message-regex "\\(.\\)")))
	  nil
	(setq mark (mew-match 2)) ;; regex includes \( \)
	(cond 
	 ((string-equal " " mark) nil)
	 (t (string-to-char mark))
	 )
	)
      )))

(defun mew-summary-mark-as (mark &optional force)
  "Mark this message if possible"
  (let ((buffer-read-only nil))
    (cond 
     ((or force (null (mew-summary-marked-p)))
      (beginning-of-line)
      (re-search-forward mew-summary-message-regex)
      (delete-char 1)
      (insert (char-to-string mark))
      (mew-highlight-mark-line mark))
     (t (message "Already marked")))
    ))

;;
;; Entire buffer
;;

(defun mew-summary-mark-exist-p (marklist)
  "See if this Summary mode has one or more marked messages"
  (let ((regex (concat
		mew-summary-message-regex
		"["
		(mapconcat (function char-to-string) marklist nil)
		"]")))
    (save-excursion
      (goto-char (point-min))
      (re-search-forward regex nil t)
      )
    ))

(defun mew-summary-mark-collect (mark &optional begin end)
  (save-excursion
    (let ((re (format (concat mew-summary-message-regex "%s") 
		      (regexp-quote (char-to-string mark))))
	  (msglist nil))
      (goto-char (if begin begin (point-min)))
      (while (re-search-forward re end t)
	(setq msglist (cons (mew-summary-message-number) msglist)))
      (nreverse msglist))))

(defun mew-summary-mark-collect2 (mark)
  (save-excursion
    (let ((re (format (concat mew-summary-message-regex "%s") 
		      (regexp-quote (char-to-string mark))))
	  (msglist nil))
      (goto-char (point-min))
      (while (re-search-forward re nil t)
	(setq msglist (cons 
		       (cons
			(mew-summary-folder-name)
			(mew-summary-message-number))
		       msglist)))
      (nreverse msglist) ;; return
      )))

;;
;; Weak marks
;;

(defun mew-summary-mark-weak (mark)
  (save-excursion
    (while (looking-at mew-summary-part-regex)
      (forward-line -1))
    (let ((cmark (mew-summary-get-mark))
	  (msg (mew-summary-message-number)))
      (cond 
       ((null msg) (message "No message"))
       ((or (equal cmark mew-mark-refile) (equal cmark mew-mark-delete))
	(message "Already strongly marked as %s" (char-to-string cmark)))
       ((equal cmark mark)
	(message "Already marked as %s" (char-to-string cmark)))
       (t ;; no mark
	(save-excursion
	  (if cmark (mew-summary-undo-one))
	  (mew-summary-mark-as mark)))
       ))
    (set-buffer-modified-p nil)
    ))

;; "@" in Summary mode
(defun mew-summary-multi ()
  "Put the multi the '@' mark on this message for 'F', 'M-s', 
and 'M-t'. It can overlay the '*' mark. The cursor stays always."
  (interactive)
  (mew-summary-mark-weak mew-mark-multi)
  )

;; "*" in Summary mode
(defun mew-summary-review ()
  "Put the review the '*' mark on this message. 
Use 'N' or 'P' to jump to a message marked with '*'.
It can overlay '@'. The cursor stays always.
See also 'mo', 'md', 'mr', and 'ma'."
  (interactive)
  (mew-summary-mark-weak mew-mark-review)
  )

(defun mew-summary-down-mark (mark)
  (interactive)
  (forward-line)
  (cond 
   ((re-search-forward (concat mew-summary-message-regex
			       (regexp-quote (char-to-string mark)))
		       nil t nil)
    (beginning-of-line)
    t)
   (t 
    (forward-line -1)
    (message "No more marked message")
    nil)))

(defun mew-summary-display-review-down ()
  "Jump to the message marked with '*' below."
  (interactive)
  (if (mew-summary-down-mark mew-mark-review)
      (if mew-summary-buffer-disp-msg ;; just seed up!
	  (mew-summary-display))))

(defun mew-summary-up-mark (mark)
  (interactive)
  (cond 
   ((re-search-backward (concat mew-summary-message-regex
			       (regexp-quote (char-to-string mark)))
			nil t nil)
    t)
   (t 
    (message "No more marked message")
    nil)
   ))

(defun mew-summary-display-review-up ()
  "Jump to the message marked with '*' above."
  (interactive)
  (if (mew-summary-up-mark mew-mark-review)
      (if mew-summary-buffer-disp-msg ;; just seed up!
	  (mew-summary-display))))

;;
;; Strong marks
;;

;; Remove

(defun mew-summary-delete (count)
  "Put the delete mark(default is 'D') on this message.
This can overlay other marks. When it overlays, the cursor stays
on the message. If it marks newly, displays the next message."
  (interactive "P")
  (if (mew-summary-part-number)
      ;; go onto the current message anyway
      ;; mew-decode-syntax-delete saves excursion...
      (re-search-backward mew-summary-message-regex nil t nil))
  (mew-decode-syntax-delete)
  (if (not (numberp count))
      (mew-summary-delete-one)
    (while (> count 0)
      (mew-summary-mark-as mew-mark-delete)
      (mew-decode-syntax-delete)
      (forward-line)
      (setq count (1- count)))
    (while (< count 0)
      (mew-summary-mark-as mew-mark-delete)
      (mew-decode-syntax-delete)
      (forward-line -1)
      (setq count (1+ count)))
    ))

(defun mew-summary-delete-one ()
  (let ((mark (mew-summary-get-mark))
	(msg (mew-summary-message-number))
	(zmacs-regions nil))
    (cond
     ((null msg) (message "No message"))
     ((equal mark mew-mark-delete) (message "Already marked as delete"))
     ((equal mark mew-mark-refile)
      (if (mew-y-or-n-p "Already marked as refile. Delete it? ")
	  (progn
	    (mew-summary-undo-one)
	    (save-excursion (mew-summary-mark-as mew-mark-delete))
	    )))
     ((or (equal mark mew-mark-review) (equal mark mew-mark-multi))
      (mew-summary-undo-one)
      (save-excursion (mew-summary-mark-as mew-mark-delete)))
     (t ;; no mark
      (save-excursion (mew-summary-mark-as mew-mark-delete))
      ;; for C-x C-x
      (beginning-of-line)
      (push-mark (point) t t)
      (mew-summary-display-next)
      )
     )))

;;
;; "mx" extensions
;;

(defun mew-summary-mark-all ()
  "Put the '*' mark onto all messages which are not marked."
  (interactive)
  (let ((regex (concat mew-summary-message-regex " ")))
    (save-excursion
      (goto-char (point-min))
      (while (re-search-forward regex nil t)
	(mew-summary-mark-as mew-mark-review)
	))
    ))

(defun mew-summary-mark-regexp (regex)
  "Put the '*' mark onto Mall messages matched to a regular expression."
  (interactive "sRegexp: ")
  (save-excursion
    (goto-char (point-min))
    (while (re-search-forward regex nil t)
      (if (not (mew-summary-marked-p))
	  (mew-summary-mark-as mew-mark-review))
      )
    ))

;;
;; Exchange
;;

(defun mew-summary-exchange-mark (old-mark new-mark)
  (let ((regex (concat mew-summary-message-regex
		       (regexp-quote (char-to-string old-mark)))))
    (save-excursion
      (goto-char (point-min))
      (if (not (re-search-forward regex nil t))
	  (message "No marked messages")
	(beginning-of-line)
	(let ((buffer-read-only nil))
	  (while (re-search-forward regex nil t)
	    (delete-backward-char 1)
	    (insert (char-to-string new-mark))
	    ))
	))
    ))
   
(defun mew-summary-mark-delete ()	; * -> D
  "Put the delete mark onto all messages marked with '*'."
  (interactive)
  (mew-summary-exchange-mark mew-mark-review mew-mark-delete))

(defun mew-summary-mark-multi ()	; * -> @
  "Change the '*' mark into the '@' mark."
  (interactive)
  (mew-summary-exchange-mark mew-mark-review mew-mark-multi))

(defun mew-summary-mark-review ()	; @ -> *
  "Change the '@' mark into the '*' mark."
  (interactive)
  (mew-summary-exchange-mark mew-mark-multi mew-mark-review))

(defun mew-summary-mark-swap ()		; @ <-> *
  "Swap the '@' mark and the '*' mark."
  (interactive)
  (mew-summary-exchange-mark mew-mark-multi mew-mark-tmp)
  (mew-summary-exchange-mark mew-mark-review mew-mark-multi)
  (mew-summary-exchange-mark mew-mark-tmp mew-mark-review)
  )

;;
;; Undo
;;

(defun mew-summary-undo-one ()
  (interactive)
  (cond 
   ((mew-summary-marked-p)
    (beginning-of-line)
    (looking-at (concat mew-summary-message-regex "\\([^ +0-9]\\)"))
    (funcall (cdr (assoc (string-to-char (mew-match 2))
			 mew-mark-undo-switch)))
    )
   ((eobp) (message "No message"))
   (t (message "No mark"))
   )
  )

(defun mew-summary-undo (count)
  "Cancel the mark on this message."
  (interactive "P")
  (if (mew-summary-part-number)
      ;; go onto the current message anyway
      ;; mew-decode-syntax-delete saves excursion...
      (re-search-backward mew-summary-message-regex nil t nil))
  (mew-decode-syntax-delete)
  (save-excursion
    (if (not (numberp count))
	(mew-summary-undo-one)
      (while (> count 0)
	(mew-summary-undo-one)
	(forward-line)
	(setq count (1- count)))
      (while (< count 0)
	(mew-summary-undo-one)
	(forward-line -1)
      (setq count (1+ count)))
      )
    )
  (set-buffer-modified-p nil)
  )

(defun mew-summary-undo-all ()
  "Cancel all marks according to what you input."
  (interactive)
  (let ((char nil) (ociea cursor-in-echo-area) (regex nil))
    (unwind-protect
	(progn
	  (message "Input mark : ")
	  (setq cursor-in-echo-area t)
	  (setq char (read-char))
	  (message "Input mark : %s" (char-to-string char))
	  )
      (setq cursor-in-echo-area ociea))
    (if (null (assoc char mew-mark-undo-switch))
	(message "Mark %s is not supported" (char-to-string char))
      (mew-decode-syntax-delete)
      (message "Undoing %s ..." (char-to-string char))
      (save-excursion
	(goto-char (point-min))
	(setq regex (concat mew-summary-message-regex
			    (regexp-quote (char-to-string char))))
	(while (re-search-forward regex nil t)
	  (mew-summary-undo-one))
	)
      (message "Undoing %s ... done" (char-to-string char))
      )
    )
  (set-buffer-modified-p nil)
  )

;;"mx" extensions
(defun mew-summary-mark-undo-all ()
  "Unmark all message marked with 'o' or 'D'."
  (interactive)
  (let ((regex (concat mew-summary-message-regex "[\\*oD]")))
    (save-excursion
      (goto-char (point-min))
      (while (re-search-forward regex nil t)
	(mew-summary-undo-one))
      ))
  )

;;
;; Processing marks
;;

(defun mew-mark-delete-region (begin end)
  (let ((buffer-read-only nil)
	(m (make-marker)))
    (set-marker m end)
    (goto-char begin)
    (while (re-search-forward 
	    (concat mew-summary-message-regex "\\([^ +0-9]\\)") m t)
      (funcall (cdr (assoc (string-to-char (mew-match 2))
			   mew-mark-delete-switch)))
      )
    (set-buffer-modified-p nil)
    ))

(defun mew-summary-exec ()
  "Process marked messages. To cancel the '*' mark, use 'u' or 'U'."
  (interactive)
  (mew-summary-exec-region (point-min) (point-max))
  )

(defun mew-summary-exec-current ()
  (interactive)
  (let (beg end)
    (save-excursion
      (beginning-of-line)
      (setq beg (point))
      (end-of-line)
      (setq end (point))
      )
    (mew-summary-exec-region beg end)
    ))

(defun mew-summary-exec-region (beg end)
  (interactive "r")
  (message "Collecting marks ...")
  (condition-case nil
      (let* ((msgs (mew-summary-mark-collect mew-mark-refile beg end))
	     (dels (mew-summary-mark-collect mew-mark-delete beg end))
	     (src (buffer-name))
	     refal folder folder-list tmp msg rm-it)
	(if (not (mew-summary-exclusive-p))
	    ()
	  (if (not (or msgs dels))
	      (message "No marks")
	    ;;
	    ;; opening...
	    ;;
	    (mew-window-configure (current-buffer) 'summary)
	    (setq mew-summary-buffer-process t) ;; this must be here
	    (message "Refiling and deleting ...")
	    ;;
	    ;; refile magic
	    ;;
	    (while msgs
	      (setq msg (car msgs))
	      (setq msgs (cdr msgs))
	      (if (setq folder-list
			(cdr (assoc msg mew-summary-buffer-refile)))
		  ;; first, remove current folder from folder-list
		  (if (mew-folder-member src folder-list)
		      (save-excursion
			(setq folder-list (delete src folder-list))
			(goto-char (point-min))
			(re-search-forward (concat "^ *" msg) nil t)
			(mew-mark-unmark)
			;; if folder-list still have folders, refile to the
			;; folders using "-link" option
			(if (> (length folder-list) 0)
			    (apply (function call-process)
				   mew-prog-immv nil nil nil
				   (format "--src=%s" src)
				   "--link"
				   (append folder-list (list msg))
				   ))
			) ;; end of save-excursion
		    ;; if multi-refile, I must refile one by one. -- nom
		    (if (nth 1 folder-list)
			(apply (function call-process)
			       mew-prog-immv nil nil nil
			       (format "--src=%s" src)
			       (append folder-list (list msg)))
		      ;; else (folder-list has only one folder)
		      (setq folder (car folder-list))
		      (if (setq tmp (assoc folder refal))
			  (setq refal (cons (append tmp (list msg))
					    (delete tmp refal)))
			(setq refal (cons (list folder msg) refal))
			))))) ;; end of while
	    ;;
	    ;; refile at once
	    ;;
	    (while refal
	      (apply (function call-process)
		     mew-prog-immv nil nil nil
		     (format "--src=%s" src)
		     (car (car refal))
		     (cdr (car refal)))
	      (setq refal (cdr refal))) ;; end of while
	    (setq mew-summary-buffer-refile nil)
	    ;;
	    ;; delete at once
	    ;;
	    (if (null dels)
		()
	      (cond
	       ((equal mew-msg-rm-policy 'always)
		(setq rm-it t))
	       ((equal mew-msg-rm-policy 'trashonly)
		(if (equal src mew-trash-folder)
		    (setq rm-it t)))
	       ((equal mew-msg-rm-policy 'uselist)
		(if (mew-member-match src mew-msg-rm-folder-list)
		    (setq rm-it t)))
	       )
	      (if rm-it
		  (apply (function call-process)
			 mew-prog-imrm nil nil nil
			 (format "--src=%s" src)
			 "--quiet=on"
			 dels)
		(if (string-equal src mew-trash-folder)
		    (let ((regex (concat mew-summary-message-regex
					 (regexp-quote (char-to-string
							mew-mark-delete)))))
		      (save-excursion
			(goto-char beg)
			(while (re-search-forward regex nil t)
			  (mew-summary-undo-one))
			))
		  (apply (function call-process)
			 mew-prog-immv nil nil nil
			 (format "--src=%s" src)
			 "--link=no"
			 mew-trash-folder
			 dels)))) ;; end of if (null dels)
	    ;;
	    ;; ending...
	    ;;
	    (mew-current-set 'message nil) ;;; hack
	    (mew-refile-guess-save)
	    (let ((next t))
	      (if (mew-summary-marked-p)
		  (mew-summary-message-down t)) ;; point moves
	      (if (mew-summary-message-number)
		  (setq next (mew-summary-message-number)))
	      (mew-decode-syntax-delete)
	      (mew-mark-delete-region beg end)
	      (mew-summary-jump-message next)
	      )
	    (if mew-summary-cache-use (mew-summary-folder-cache-save))
	    (setq mew-summary-buffer-process nil)
	    (run-hooks 'mew-summary-exec-hook)
	    (mew-summary-mode-line (current-buffer))
	    (set-buffer-modified-p nil)
	    (message "Refiling and deleting ... done")
	    ))) ;; end of let
    (quit 
     (setq mew-summary-buffer-process nil)
     (set-buffer-modified-p nil)
     (message "Refiling and deleting ... imcompletely done")
     )) ;; end of condision-case
  )

;;;
;;; Clean up marks!
;;;

(defun mew-mark-clean-up (&optional sum-buf)
  "Process marked messages for this folder"
  (if sum-buf
      (set-buffer sum-buf)
    (setq sum-buf (buffer-name)))
  (if (mew-summary-mark-exist-p (list mew-mark-delete mew-mark-refile))
      (if (mew-y-or-n-p "Marks exist in %s. Process them? " sum-buf)
	  (mew-summary-exec)
	))
  )

(defun mew-mark-clean-up-all ()
  "Process marked messages for all Summary modes.
Typically called by kill-emacs-hook."
  (let ((bufs mew-clean-up-buffers))
    (save-excursion
      (while bufs
	(if (bufferp (get-buffer (car bufs)))
	    (mew-mark-clean-up (car bufs)))
	(setq bufs (cdr bufs))
	))
    ))

(provide 'mew-mark)
